package grammar;

import java.util.HashSet;

import machine.Machine;
import values.Value;
import context.TerminalSet;
import env.Env;
import exp.BoaConstructor;

@BoaConstructor(fields = { "left","right" })
public class Or extends PTerm {

	public PTerm left;
	public PTerm right;

	public Or() {
	}

	public Or(PTerm left, PTerm right, PTerm... rest) {
		super();
		this.left = left;
		this.right = consTail(right, rest,0);
	}

	PTerm consTail(PTerm tail, PTerm[] rest,int i) {
		for (PTerm term : rest)
			tail = new Or(tail, term);
		return tail;
	}

	public PTerm close(Env<String, Value> env) {
		return new Or(left.close(env), right.close(env));
	}

	public TerminalSet predictors(Env<String, Value> env, HashSet<String> NTs) {
		TerminalSet P = left.predictors(env, NTs);
		TerminalSet Q = right.predictors(env, NTs);
		TerminalSet R = new TerminalSet();
		for(String p : P) R.add(p);
		for(String q : Q) R.add(q);
		return R;
	}

	public void exec(Machine machine) {
		machine.choice(right);
		machine.pushInstr(left);
	}

	public String pprint(int opPrec) {
		if(opPrec < OROP) 
			return "(" + left.pprint(OROP) + " | " + right.pprint(OROP) + ")";
		else return left.pprint(OROP) + " | " + right.pprint(OROP); 
	}

}
