package xpl;

import machine.Machine;
import values.Record;
import values.Value;

import commands.Command;
import commands.Quit;

import context.Context;
import context.TopLevel;
import env.Env;
import exp.Exp;
import exp.Import;
import grammar.Call;
import grammar.Grammar;

public class Interpreter {

	public static void main(String[] args) {
		Env<String, Value> env = Value.builtinEnv;
		Machine machine = new Machine(null, env, new TopLevel(System.in), 0, null, 0);
		XPL.XPL = (Grammar) XPL.XPL.eval(machine);
		Record xpl = Import.getFile("src/xpl/xpl.xpl");
		XPL.setXPL((Grammar)xpl.dot("XPL"));
		boolean quit = false;
		while (!quit) {
			try {
				Context.reset();
				machine = new Machine(XPL.XPL, env, new TopLevel(System.in), 0, null, 0);
				System.out.print("> ");
				machine.pushInstr(new Call("topLevelCommand"));
				Command command = (Command) machine.run();
				if (machine.isOk()) {
					if (command instanceof Quit)
						quit = true;
					else if (command instanceof commands.Import) {
						commands.Import i = (commands.Import) command;
						for (Exp exp : i.exps) {
							Value value = exp.eval(machine);
							values.Record record = Import.getRecord(value);
							for (values.Field field : record.fields)
								machine.bind(field.name, field.value);
						}
					} else if (command instanceof commands.Exp) {
						commands.Exp exp = (commands.Exp) command;
						System.out.println(exp.exp.eval(machine));
					} else if (command instanceof commands.Bind) {
						commands.Bind bind = (commands.Bind) command;
						Value value = bind.value.eval(machine);
						machine.bind(bind.name, value);
						System.out.println(bind.name + " => " + value);
					}
				} else
					System.out.println(machine.getError());
				env = machine.getEnv();
			} catch (Error e) {
				System.out.println(e.getMessage());
			}
		}
		System.out.println("...terminated");
	}

}
